/**
 * {文件操作的一些API}
 * @author yutent<yutent.io@gmail.com>
 * @date 2023/11/06 15:46:27
 */

// 计算文件指纹
// 为避免大文件耗时过长, 只取前后各64位, 计算sha1值
export function getFingerprint(file) {
  let head = file.slice(0, 64)
  let tail = file.slice(-64)
  return Promise.all([head.arrayBuffer(), tail.arrayBuffer()]).then(
    ([a1, a2]) => {
      let u8a1 = new Uint8Array(a1)
      let u8a2 = new Uint8Array(a2)
      let u8 = new Uint8Array(u8a1.length + u8a2.length)
      u8.set(u8a1)
      u8.set(u8a2, u8a1.length)
      return u8
    }
  )
}

export function toBlobURL(url, type) {
  return fetch(url)
    .then(r => r.blob())
    .then(r => URL.createObjectURL(type ? new Blob([r], { type }) : r))
}

export function parseSize(num) {
  if (num < 1024) {
    return `${num} B`
  } else {
    num = (num / 1024).toFixed(2) - 0

    if (num < 1024) {
      return `${num} KB`
    } else {
      num = (num / 1024).toFixed(2) - 0
      return `${num} MB`
    }
  }
}

export async function toUint8(bin) {
  return new Uint8Array(await bin.arrayBuffer())
}

// 路径合并, 同nodejs的 path.join
export function join(p1, p2) {
  let tmp1 = p1.split('/')
  let tmp2 = p2.split('/')
  if (tmp1.at(-1) === '') {
    tmp1.pop()
  }
  while (tmp2.length) {
    let tmp = tmp2.shift()
    if (tmp === '.' || tmp === '') {
      continue
    } else if (tmp === '..') {
      tmp1.pop()
    } else {
      tmp1.push(tmp)
    }
  }
  return tmp1.join('/')
}

// 获取路径的文件名
export function filename(path) {
  let paths = path.split('/')
  return paths.pop()
}

async function getEntry(handle, dir = '') {
  let key = join(dir, handle.name)
  if (handle.kind === 'directory') {
    let entries = await handle.entries()
    let item
    let list = []

    while ((item = await entries.next())) {
      if (item.done) {
        break
      }
      let tmp = item.value[1]

      if (tmp.kind === 'directory') {
        list = list.concat(await getEntry(tmp, key))
      } else {
        list.push({
          key: join(key, tmp.name),
          file: await tmp.getFile()
        })
      }
    }
    return list
  } else {
    return {
      key,
      file: await handle.getFile()
    }
  }
}

export async function resolveFiles(items) {
  // 无法直接在 getAsFileSystemHandle 这一步就用 await
  // 会导致 dataTransfer 后面的数据被清空, 无论选中多少个, 都只能读到第一个
  return Promise.all([...items].map(it => it.getAsFileSystemHandle())).then(
    async handles => {
      let list = []
      for (let it of handles) {
        list = list.concat(await getEntry(it))
      }
      return list
    }
  )
}
